/**
 * 全局http请求的promise对象封装;
 * ### 注意使用必须加上错误处理 ###
 * 参数设置请参考
 * @link https://github.com/github/fetch#post-json
 * @link https://developer.mozilla.org/zh-CN/docs/Web/API/Fetch_API/Using_Fetch
 * @link https://github.com/camsong/blog/issues/2
 * @module util/invokeApi
 */
import 'whatwg-fetch'; // fetch相关polyfill库由'isomorphic-fetch'库更换过来
import path from 'lodash/fp/path';
import { httpUrlPrefixer, APIversion } from '../../config';
import { storeLogin, clearLogin } from './auth';

function queryParams(params) {
  return Object.keys(params)
    .map(k => `${encodeURIComponent(k)}=${encodeURIComponent(params[k])}`)
    .join('&');
}

/**
 * @config {
 *  contentType: {String}请求数据是否是'json'
 *  acceptType: {String} 接收数据的格式'json','blob'等
 * }
 * url {String}   - 请求url
 * {Object} options - 请求参数
 * {*|Promise.<TResult>} - 请求的promise对象
 */
const invokeApi = config => (options = { headers: {} }) => (uri) => {
  let url = uri;
  const opt = { ...options };
  if (opt.version) {
    url = `/${opt.version}${url}`;
  } else {
    url = `/${APIversion}${url}`;
  }
  if (options.params) {
    url += (uri.indexOf('?') === -1 ? '?' : '&') + queryParams(options.params);
    delete opt.params;
  }
  if (options.body !== undefined && config.contentType === 'json') {
    opt.body = JSON.stringify(options.body);
  }
  if (typeof options.headers === 'undefined') {
    opt.headers = {};
  }
  if (url !== `/${APIversion}/auth/login` && url !== `/${APIversion}/auth/ssologin` && url !== `/${APIversion}/auth/refreshtoken`) {
    if (typeof options.headers === 'undefined' || typeof options.headers.Authorization === 'undefined') {
      opt.headers.Authorization = `${sessionStorage.tokenType} ${sessionStorage.token}`;
    }
  }

  if (path(['headers', 'Content-Type'])(options) === undefined && config.contentType === 'json') {
    Object.assign(opt.headers, {
      'Content-Type': 'application/json',
    });
  }

  const fetchPromise = fetch(`${httpUrlPrefixer}${url}`, opt)
    .then((response) => {
      if (response.status === 401 && localStorage.refreshToken && sessionStorage.loginType) {
        return fetch(`${httpUrlPrefixer}/${APIversion}/auth/refreshtoken`, {
          headers: {
            'Content-Type': 'application/json',
            refreshToken: String(localStorage.refreshToken),
          },
        }).then((c) => {
          if (!c.ok) {
            /**
             * refresh失效，重新刷面
             */
            clearLogin();
            location.href = '/';
            return Promise.reject(c);
          }
          return c.json();
        }).then((ch) => {
          if (ch.resultCode === '000000') {
            storeLogin()(ch.tokenInfo);
            opt.headers.Authorization = `${sessionStorage.tokenType} ${sessionStorage.token}`;
            return fetch(`${httpUrlPrefixer}${url}`, opt)
              .then((chunk) => {
                if (!chunk.ok) {
                  return Promise.reject(chunk);
                }
                switch (config.acceptType) {
                  case 'json':
                    return chunk.json();
                  case 'blob':
                    return chunk;
                  case 'text':
                    return chunk.text();
                  case 'arrayBuffer':
                    return chunk.arrayBuffer();
                  default:
                    return chunk.json();
                }
              });
          }
          clearLogin();
          // 重新刷面
          location.href = '/';
          return Promise.reject(ch);
        });
      } else if (response.status === 401 && sessionStorage.token === undefined) {
        // 接口返回 401 并且 localStorage token 不存在 认定为用户清除了浏览器历史记录 需重新登录
        clearLogin();
        // 重新刷面
        location.href = '/';
      } else if (!response.ok) {
        return Promise.reject(response);
      }

      switch (config.acceptType) {
        case 'json':
          return response.json();
        case 'blob':
          return response;
        case 'text':
          return response.text();
        case 'arrayBuffer':
          return response.arrayBuffer();
        default:
          return response.json();
      }
    });
  if (options.timeout) {
    /**
     * 超时参数设置为options上的timeout参数，类型Number，单位ms；
     */
    let abortFunc = null;
    const abortPromise = new Promise((resolve, reject) => {
      abortFunc = () => {
        reject(new Error('request timeout'));
      };
    });
    const abortablePromise = Promise.race([
      fetchPromise, abortPromise,
    ]);
    setTimeout(() => {
      abortFunc();
    }, options.timeout);
    return abortablePromise;
  }
  return fetchPromise;
};

export const sendBlobApi = invokeApi({ contentType: '', acceptType: 'json' });

export const getBlobApi = invokeApi({ contentType: 'json', acceptType: 'blob' });

export const getApi = invokeApi({ contentType: 'json', acceptType: 'json' })();

export default invokeApi({ contentType: 'json', acceptType: 'json' });
